#!/usr/bin/env python

"""
Translate the AST of a Python program into a more interpretable representation.

Copyright (C) 2007 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import micropython.inspect
from micropython.rsvp import *
import compiler.ast
from compiler.visitor import ASTVisitor
try:
    set
except NameError:
    from sets import Set as set

class TranslateError(Exception):

    "A translation error."

    pass

# Program visitors.

class Translation(ASTVisitor):

    "A translated module."

    def __init__(self, module):

        "Initialise the translation with an inspected 'module'."

        ASTVisitor.__init__(self)
        self.visitor = self
        self.module = module
        self.unit = None

    def get_module_code(self):

        "Return the top-level module code."

        self.unit = self.module
        return self.dispatch(self.module.module)

    def get_code(self, unit):

        "Return the code for the given 'unit'."

        self.unit = unit
        return self.dispatch(unit.node)

    def __repr__(self):
        return "Translation(%r)" % self.module

    def get_scope(self, name):
        if self.unit.has_key(name):
            return "local"
        elif self.module.has_key(name):
            return "global"
        else:
            return "builtins"

    # Visitor methods.

    def default(self, node, *args):
        raise TranslateError, node.__class__

    def dispatch(self, node, *args):
        return ASTVisitor.dispatch(self, node, *args)

    def _visitName(self, node, classes):
        name = node.name
        scope = self.get_scope(name)
        NameInstruction, AttrInstruction = classes

        if scope == "local":
            unit = self.unit
            if isinstance(unit, micropython.inspect.Function):
                return [NameInstruction(unit.locals()[name])]
            elif isinstance(unit, micropython.inspect.Class):
                return [AttrInstruction(unit.all_class_attributes()[name])]
            elif isinstance(unit, micropython.inspect.Module):
                return [AttrInstruction(unit.module_attributes()[name])]
            else:
                raise TranslateError, "Program unit %r has no local %r" % (unit, name)

        elif scope == "global":
            globals = self.module.module_attributes()
            if globals.has_key(name):
                return [AttrInstruction(globals[name])]
            else:
                raise TranslateError, "Module %r has no attribute %r" % (self.module, name)

        else:
            builtins = micropython.inspect.builtins.module_attributes()
            return [AttrInstruction(builtins[name])]

    def visitAdd(self, node): return []

    """
    _t1 = node.left
    _t2 = node.right
    try:
        _t1.__add__(_t2)
    except AttributeError:
        _t2.__radd__(_t1)
    """

    def visitAnd(self, node): return []

    def visitAssert(self, node): return []

    def visitAssign(self, node):
        results = []
        results += self.dispatch(node.expr)
        for n in node.nodes:
            results += self.dispatch(n)
        return results

    def visitAssAttr(self, node): return []

    def visitAssList(self, node): return []

    def visitAssName(self, node):
        return self._visitName(node, (StoreName, StoreAttr))

    visitAssTuple = visitAssList

    def visitAugAssign(self, node): return []

    def visitBackquote(self, node): return []

    def visitBitand(self, node): return []

    def visitBitor(self, node): return []

    def visitBitxor(self, node): return []

    def visitBreak(self, node): return []

    def visitCallFunc(self, node): return []

    def visitClass(self, node): return []

    def visitCompare(self, node): return []

    def visitConst(self, node): return []

    def visitContinue(self, node): return []

    def visitDecorators(self, node): return []

    def visitDict(self, node): return []

    def visitDiscard(self, node):
        return self.dispatch(node.expr)

    def visitDiv(self, node): return []

    def visitEllipsis(self, node): return []

    def visitExec(self, node): return []

    def visitExpression(self, node): return []

    def visitFloorDiv(self, node): return []

    def visitFor(self, node): return []

    def visitFrom(self, node): return []

    def visitFunction(self, node):

        # Only store the name when visiting this node from outside.

        if self.unit is self.module:
            return self._visitName(node, (StoreName, StoreAttr))
        else:
            return self.dispatch(node.code)

    def visitGenExpr(self, node): return []

    def visitGenExprFor(self, node): return []

    def visitGenExprIf(self, node): return []

    def visitGenExprInner(self, node): return []

    def visitGetattr(self, node): return []

    def visitGlobal(self, node): return []

    def visitIf(self, node):
        for test, body in node.tests:
            self.dispatch(body)
        if node.else_ is not None:
            self.dispatch(node.else_)
        return None

    def visitImport(self, node): return []

    def visitInvert(self, node): return []

    def visitKeyword(self, node): return []

    def visitLambda(self, node): return []

    def visitLeftShift(self, node): return []

    def visitList(self, node): return []

    def visitListComp(self, node): return []

    def visitListCompFor(self, node): return []

    def visitListCompIf(self, node): return []

    def visitMod(self, node): return []

    def visitModule(self, node):
        return self.dispatch(node.node)

    def visitMul(self, node): return []

    def visitName(self, node):
        return self._visitName(node, (LoadName, LoadAttr))

    def visitNot(self, node): return []

    def visitOr(self, node): return []

    def visitPass(self, node): return []

    def visitPower(self, node): return []

    def visitPrint(self, node): return []

    def visitPrintnl(self, node): return []

    def visitRaise(self, node): return []

    def visitReturn(self, node): return []

    def visitRightShift(self, node): return []

    def visitSlice(self, node): return []

    def visitStmt(self, node):
        result = []
        for n in node.nodes:
            result += self.dispatch(n)
        return result

    def visitSub(self, node): return []

    def visitSubscript(self, node): return []

    def visitTryExcept(self, node): return []

    def visitTryFinally(self, node): return []

    def visitTuple(self, node): return []

    def visitUnaryAdd(self, node): return []

    def visitUnarySub(self, node): return []

    def visitWhile(self, node): return []

    def visitWith(self, node): return []

    def visitYield(self, node): return []

# vim: tabstop=4 expandtab shiftwidth=4
